{-# LANGUAGE Arrows #-}

import Data.Maybe(fromMaybe,listToMaybe)
import Text.XML.HXT.Arrow
import qualified Text.XML.HXT.DOM.XmlNode as X
import System.Environment(getArgs)
import Text.ParserCombinators.Parsec(GenParser,(<|>),incSourceColumn,many,parse,tokenPrim)

import Raw

main =
    do fs <- getArgs
       case fs of
         [] -> putStrLn ("xmlrender SRC [DEST]")
         f:gs -> do
           let g = case gs of {h:_ -> h; _ -> ""}
           runX (m f g)
           return ()
    where
      m f g = Raw.readXmlDoc [] f >>> render >>>
              ifA hasXmlPi (writeDocument [] g)
                           (writeDocument [(a_no_xml_pi,v_yes)] g)

node :: (ArrowXml a) => String -> a XmlTree XmlTree
node s = isElem >>> hasName s

{- renders article using a class (template).
   The class is dictated by the class attribute of the article,
   i.e., the article should look like:

     <article class="mytemplate.xml"> ... </article>

   and the template is read from the file mytemplate.xml.
   The rendering proceeds by substituting HOLE's in the template
   by elements in the article.
-}
render =
    proc article_root -> do
      article <- getChildren >>> node "article" >>> dotable >>> domath -< article_root
      cls <- getAttrValue0 "class" -< article
      template <- Raw.readXmlFrom [] -< cls
      processTopDownUntil (substWith article) -<< template

{- if the current element (from the template) is to be substituted
   (name is HOLE),
   do so by returning the matching child node from article.
-}
substWith article =
    proc element -> do
      n <- getAttrValue0 "using" <<< node "HOLE" -< element
      t <- getAttrValue "transform" -< element
      (getChildren <<< transform t <<< node n <<< getChildren) `orElse` txt "" -<< article

{- possible transformations during substitution -}
transform "textonly" = canonicalizeContents >>>
                       replaceChildren (deep isText) >>>
                       collapseXText
transform _ = this

{- table translation -}

dotable = processTopDown (tabulate `when` node "tab") where
  tabulate = mkelem "table" [getAttrl] [listA getChildren >>> arrL convert]
  convert s = case parse tablestuff "tab" s of
    Left e -> [X.mkCmt (show e)]
    Right s -> s

tablestuff :: GenParser XmlTree () [XmlTree]
tablestuff = do
    s <- many other
    r <- many tr
    return (s ++ [mknode "tbody" [] r])

tr = do
    e <- tokenSat (node "r")
    many other
    s <- many (th <|> td)
    return (mknode "tr" (attrs e) s)

td = do
    e <- tokenSat (node "d")
    s <- many other
    return (mknode "td" (attrs e) s)

th = do
    e <- tokenSat (node "h")
    s <- many other
    return (mknode "th" (attrs e) s)

other = tokenSat (neg (node "r" <+> node "d" <+> node "h"))

tokenSat p = tokenPrim show (\pos _ _ -> incSourceColumn pos 1) (listToMaybe . runLA p)

mknode name a s = X.mkElement (mkName name) a s

attrs e = runLA (getAttrl >>> changeQName edit) e where
  edit n = mkName (subst replacements (localPart n))

replacements = [("c","class"),("x","colspan"),("y","rowspan")]

subst s x = fromMaybe x (lookup x s)


{- math translation: configuration -}

isInline = node "m"
isIntext = node "t"
isDisplay = node "math"
isNest = node "nest"
isReserved = node "reserved"

varbegin = ['A'..'Z'] ++ ['a'..'z']
var = "'_" ++ ['0'..'9'] ++ varbegin

mkvar = mkelem "span" [sattr "class" "mathvar"] [mkText]
mkres = mkelem "span" [sattr "class" "mathres"] [mkText]

{- math translation: algorithm -}

domath = proc article -> do
    b <- processTopDownUntil (isInline<+>isDisplay<+>isReserved >>> canonicalizeContents) -< article
    reserved <- listA (getChildren >>> isReserved /> getText >>> arrL words) -< b
    rewrite reserved False -<< b

{- r: list of reserved words; m: math mode true/false -}
rewrite r m = choiceA
  [ isText :-> (annotate r `whenP` (const m)) ,
    isInline :-> (getChildren >>> rewrite r True) ,
    isIntext :-> (getChildren >>> rewrite r False) ,
    isDisplay :-> mkelem "div" [editclass] [getChildren >>> rewrite r True] ,
    isNest :-> mkelem "div" [editclass] [getChildren >>> rewrite r m] ,
    this :-> processChildren (rewrite r m)
  ] where
    editclass = addAttrl (attr "class" (getAttrValue "class" >>> arr addnest >>> mkText)) >>> getAttrl
    addnest s = let w = words s
                in unwords (if "nest" `elem` w then w else "nest":w)

annotate r = getText >>> arrL identify >>>
             (ifP (`elem` r) mkres mkvar) ||| mkText

{- identify "num = len + 1"
   ->
   [Left"num", Right" = ", Left"len", Right" + 1"]
-}
identify "" = []
identify (c:cs) | c `elem` varbegin = let (vs,ds) = span (`elem` var) cs
                                      in Left (c:vs) : identify ds
                | otherwise = let (ds,vs) = break (`elem` varbegin) cs
                              in  Right (c:ds) : identify vs
